########
# (c) Thomas A. Adams II
# December 2020
# From Bonus Tutorial 2 in the book
# LEARN ASPEN PLUS IN 24 HOURS, McGraw-Hill Education
#
# This code does Latin Hypercube Sampling of an example Aspen Plus simulation
# In particular, it selects many temperature and pressure combinations of the reactor in the simulation
# Then it simulates them all using parallel processing, and prints out the results of two key variables
#
# It has no error checking, and is for demonstration purposes only. It may fail on your computer with large runs in Aspen Plus v12 and earlier.
# If it fails, some of the points would be missing from the results.
# See Part2b.py for the robust version which will overcome this problem on most Windows PCs.
#
# USER PARAMETERS TO SET
# Users should change these two numbers found near the top of the main function:
#
#   NumAspenCopiesOpen = 4  --> This should be between 1 and the number of processors your computer has, but no more than 60     
#   PointsPerProcessor = 10 --> This is how many data points you want *per* processor. The total number of points explored is NumAspenCopiesOpen * PointsPerProcessor 
#
# HOW TO RUN
# In the windows command prompt, type: 
#
#     python Part2.py
#
# If you want to save the results to a file instead of seeing them on the screen, instead type 
#
#     python Part2.py > filename.txt
#
# USEFUL FUNCTIONS
# CRTL + C or CTRL + Break during the run will stop it, for example, if it is taking too long. Expect betwen 0.15 and 0.5 seconds per point visited. 
# If you break it early or something goes wrong, you may have many open copies of Aspen Plus in background. If it is too tedious to close them with the Windows Task Manager you can use the following command prompt command:
#
#     taskkill /IM "AspenPlus.exe" /f   
#
# If you find there are too many temporary Aspen plus files lying about (those that start with underscore) you can type this at the command prompt:
#
#     del _*.*
#
# Just be sure to include that _ or else you will delete all files!!!
#
########

import os                            # This connects to your operating system   
import win32com.client as Windows    # This is a package that lets you call Windows programs
import multiprocessing as mp         # This is the multiprocessing package
import random                        # This lets us call the random number generator
import time                          # This has the timer function in it


# This runs the Aspen Sim for the given T and P, and returns the result. Does not do error checking.
def RunAspenPlusReactor(JobID, Points):
    
    # Create the Aspen Plus Object     
    AspenPlus = Windows.Dispatch('Apwn.Document')
    
    # Open the file    
    AspenPlus.InitFromArchive2(os.path.abspath('Tutorial B02.bkp'))
    
    # Now, run a simulation for each point we were given
    for point in range(len(Points)):                            
                                    
        # Set the T and P variables for the current point as the T and P of the reactor in the simulation   
        AspenPlus.Tree.FindNode('\Data\Blocks\REACTOR\Input\TEMP').Value = Points[point][0] # Temperature in C
        AspenPlus.Tree.FindNode('\Data\Blocks\REACTOR\Input\PRES').Value = Points[point][1] # Pressure in bar                         

        # Run the simulation                                                       
        AspenPlus.Engine.Run2()
        
        # Get the results
        Q = AspenPlus.Tree.FindNode('\Data\Blocks\REACTOR\Output\QCALC').Value       # The reactor heat duty in Gcal/hr
        CO = AspenPlus.Tree.FindNode('\Data\Streams\PRODUCTS\Output\MOLEFRAC\MIXED\CO').Value # The mole fraction of CO
        
        # Pint out the results
        print(f"In JobID: {JobID} T: {Points[point][0]:0.2f} P: {Points[point][1]:0.2f} Q: {Q:0.4f} CO: {CO:0.4f}")                    
                   
    return (JobID)                
        
if __name__ == '__main__': 
    
    #########################
    # USER CHANGE THIS PART #
    #########################
    NumAspenCopiesOpen = 4     
    PointsPerProcessor = 10
    #########################
    #########################
    
    NPoints = NumAspenCopiesOpen * PointsPerProcessor
    
    # Create some empty lists for us to fill in later
    PointsList = list()
    T = list()
    P = list()
    
    # Starts the timer
    TimeStart = time.perf_counter()
    
    # Create the LHS points
    for i in range(NPoints):              
        T.append(600 + (400) * (i/(NPoints-1)))
        P.append(  2 +  (50) * (i/(NPoints-1)))        
    random.shuffle(P)     
    
    # Make job lists for each processor
    for i in range(NumAspenCopiesOpen):
        for c in range(PointsPerProcessor):
            if (1 > c):
                PointsList.append(list())
            PointsList[i].append([T[c + i * PointsPerProcessor], P[c + i * PointsPerProcessor]])   
                           
    # Let's see how many processors we have (this is logical, not physical)
    print("Number of processors available: ", mp.cpu_count())
    print("Number of processes used: ", NumAspenCopiesOpen)
                                                      
    # Create the pool of processors. Each processor will have one copy of Aspen Plus to work with
    pool = mp.Pool(NumAspenCopiesOpen) # This creates a pool of workers using all of the available ones.
    
    # Start all the processors!             
    ResultData = [pool.apply_async(RunAspenPlusReactor, args=(JobID, PointsList[JobID])) for JobID in range(NumAspenCopiesOpen)]    
    
    # Closes the parallel worker pool        
    pool.close()

    # This waits until all the above jobs are finished before continuing on
    pool.join()        
                   
    # Stop the timer. We are done!                   
    TimeFinish = time.perf_counter()            
    print(f"Finished in {TimeFinish - TimeStart:0.2f} Seconds.")
        



