########
# (c) Thomas A. Adams II
# December 2020
# From Bonus Tutorial 2 in the book
# LEARN ASPEN PLUS IN 24 HOURS, McGraw-Hill Education
#
# This code does Latin Hypercube Sampling of an example Aspen Plus simulation
# In particular, it selects many temperature and pressure combinations of the reactor in the simulation
# Then it simulates them all using parallel processing, and prints out the results of two key variables
#
# This version has error checking. It uses the following strategy: 
#   Each process has one copy of Aspen Plus open
#   Each simulation run uses the results of the previous run as the initial guesses, so no reinitilzation is used.
#   If any errors occurred, it reinitializes and tries again
#   If still any errors occurred after reinitializing, it closes Aspen PLus and opens a new copy and tries again
#   If that still doesn't work, it records a 0 for all variables, gives up, and moves on. 
#
# In our tests this strategy is very effective, fast, and robust. 
# Note that the final "give up" state happened extremely infrequently in our test cases, and only on certain machines
# If you use a differenet Aspen Plus simulation that is more difficult to converge, you may have this far more often, however.     
#
# USER PARAMETERS TO SET
# Users should change these two numbers found near the top of the main function:
#
#   NumAspenCopiesOpen = 4  --> This should be between 1 and the number of processors your computer has, but no more than 60     
#   PointsPerProcessor = 10 --> This is how many data points you want *per* processor. The total number of points explored is NumAspenCopiesOpen * PointsPerProcessor 
#
# HOW TO RUN
# In the windows command prompt, type: 
#
#     python Part2b.py
#
# If you want to save the results to a file instead of seeing them on the screen, instead type 
#
#     python Part2b.py > filename.txt
#
# USEFUL FUNCTIONS
# CRTL + C or CTRL + Break during the run will stop it, for example, if it is taking too long. Expect betwen 0.15 and 0.5 seconds per point visited. 
# If you break it early or something goes wrong, you may have many open copies of Aspen Plus in background. If it is too tedious to close them with the Windows Task Manager you can use the following command prompt command:
#
#     taskkill /IM "AspenPlus.exe" /f   
#
# If you find there are too many temporary Aspen plus files lying about (those that start with underscore) you can type this at the command prompt:
#
#     del _*.*
#
# Just be sure to include that _ or else you will delete all files!!!
#
########

import os                            # This connects to your operating system, and lets you make operating system calls, like get paths to directories, change directories, etc   
import win32com.client as Windows    # This is a package that lets you call Windows programs
import multiprocessing as mp         # This is the multiprocessing package
import random                        # This lets us call the random number generator
import time                          # This has the timer function in it


# This runs the Aspen Sim for the given T and P, and returns the result. Does not do error checking.
def RunAspenPlusReactor(JobID, Points):
    
    # Create the Aspen Plus Object     
    AspenPlus = Windows.Dispatch('Apwn.Document')
    
    # These are variables that keep track of how many times certain errors happened 
    reinits = 0
    redispatches = 0
    varfails = 0
    finalfails = 0    
    
    # Load the simulation from the file, and fail if it doesn't work the first time
    # If it doesn't work the first time, usually this means the file name is wrong, or the license checkout failed.
    try: 
        AspenPlus.InitFromArchive2(os.path.abspath('Tutorial B02.bkp'))
    except:
        print(f"Could not load the file #{JobID}. Ensure that it is in the same directory as the python script. Also, ensure that you are connected to a license server and that you have Aspen Plus licenses available.")
        return (None)
    
    # Create the empty list for storing results
    Data = list()

    # Now, run a simulation for each point we were given
    for point in range(len(Points)):
            
        # This keeps track of how many times we attempted to run the simulation for this point 
        attempt = 0
        
        # Default values for the results are stored here, in case of failure
        Q = 0
        CO = 0
        
        # Run The Simulation. We allow 3 attempts before giving up.
        while (3 > attempt):
            
            # If our first attempt failed, try re-initializaing the simulation and try again
            # This could be just because the initial guess (the previous run) was poor
            if (1 == attempt):                
                print("Attempting to reinitialize ", JobID)
                try:
                    # Reinitialize the simulation
                    AspenPlus.Engine.Reinit()
                    reinits = reinits + 1
                except:
                    print("Could not reinitialize ", JobID)
                    attempt = attempt + 1                    
                else: 
                    print("Reinitialize successful ", JobID)                   
            
            # If our second attempt failed, try relaunching Aspen Plus entirely             
            if (2 == attempt):                
                print("Attempting to redispatch Aspen Plus", JobID)                                
                try:
                    AspenPlus = Windows.Dispatch('Apwn.Document')
                    redispatches = redispatches + 1
                except:
                    print("Could not dispatch on reopen ", JobID)
                    break     
                else:
                    print("Dispatch successful ", JobID)                    
                
                # At this point, the dispatch worked, so try to open the file again.                 
                try:                                 
                    AspenPlus.InitFromArchive2(os.path.abspath('Tutorial B02.bkp'))
                except:
                    print("Could not initialize on reopen ", JobID)
                    break                     
                else:
                    print("Initialize on reopen successful ", JobID)
            
            # At this point, we have an open and working file.
            # Set the T and P variables for the current point as the T and P of the reactor in the simulation                                   
            AspenPlus.Tree.FindNode('\Data\Blocks\REACTOR\Input\TEMP').Value = Points[point][0] # Temperature in Celcius
            AspenPlus.Tree.FindNode('\Data\Blocks\REACTOR\Input\PRES').Value = Points[point][1] # Pressure in bar                        
            
            # Run the simulation                                           
            try:
                AspenPlus.Engine.Run2()
            except:
                # If it did not run, print an error message
                print("Aspen Run Error in jobID ", JobID, " on attempt" , attempt)
                # Increment the attempt counter, so can try again in the next while loop iteration
                attempt = attempt + 1                                    
            else:
                # The simulation was successful. Let's record that it worked this time for later.
                worked = 1                            
                
                # Sometimes the simulation can be successful but some of the simulation varibles are missing
                # We need to check for that here.
                try:                
                    Q = AspenPlus.Tree.FindNode('\Data\Blocks\REACTOR\Output\QCALC').Value
                except:
                    # Even though this simulation worked, there is some problem in Aspen Plus that prevents us 
                    # From seeing the variable. So We can set a default variable and indicate that it did not work.
                    Q = 0
                    worked = 0
                    varfails = varfails + 1
                    
                # We need to check against each variable that we want to keep. A little tedious but important for robustness     
                try: 
                    CO = AspenPlus.Tree.FindNode('\Data\Streams\PRODUCTS\Output\MOLEFRAC\MIXED\CO').Value
                except:
                    CO = 0
                    worked = 0
                    varfails = varfails + 1
                    
                # If none of the variable pulls failed, then we can break from our while loop because the point was successful     
                if (1 == worked):                
                    break
                else:
                    # Otherwise we just try again
                    attempt = attempt + 1    
         
        # This checks to see if the previous while loop exited with a simulation run and all output variables obtained.
        # If not, we just keep track of the number of times this happened for convenience.             
        if (2 < attempt):
            finalfails = finalfails + 1                                            
        
        # Store and print the data. Although I am not doing anything with the Data object, you could do something with it if you wanted.        
        Data.append([Q, CO])                       
        print(f" {Points[point][0]:0.4f}, {Points[point][1]:0.4f}, {Q:0.4f}, {CO:0.4f}; ")
    
    # Close this out and spit out the data.  
    print("JobID: ", JobID, "Finished with this worker. Completed ", len(Data), " points, with ", reinits, " reinitializes, ", redispatches, " redispatches, ", varfails, " variable pull fails, and ", finalfails, " complete failures.")           
    return (JobID, Data)                
        
if __name__ == '__main__': 
     
    #########################
    # USER CHANGE THIS PART #
    #########################
    NumAspenCopiesOpen = 4     
    PointsPerProcessor = 300
    #########################
    #########################
    
    NPoints = NumAspenCopiesOpen * PointsPerProcessor
    
    # Create some empty lists for us to fill in later
    PointsList = list()
    T = list()
    P = list()
    
    # Starts the timer
    TimeStart = time.perf_counter()
    
    # Create the LHS points
    for i in range(NPoints):              
        T.append(600 + (200) * (i/(NPoints-1)))
        P.append(  2 +  (50) * (i/(NPoints-1)))        
    random.shuffle(P)     
    
    # Make job lists for each processor
    for i in range(NumAspenCopiesOpen):
        for c in range(PointsPerProcessor):
            if (1 > c):
                PointsList.append(list())
            PointsList[i].append([T[c + i * PointsPerProcessor], P[c + i * PointsPerProcessor]]) 
            
    # Check to make sure that the user did not specify the wrong number of processors
    if (0 > NumAspenCopiesOpen or mp.cpu_count() < NumAspenCopiesOpen):
        NumAspenCopiesOpen = mp.cpu_count()
    if (60 < NumAspenCopiesOpen):
        NumAspenCopiesOpen = 60 
                           
    # Let's see how many processors we have (this is logical, not physical)
    print("Number of processors available: ", mp.cpu_count())
    print("Number of processes used: ", NumAspenCopiesOpen)
                                                      
    # Create the pool of processors. Each processor will have one copy of Aspen Plus to work with
    pool = mp.Pool(NumAspenCopiesOpen) # This creates a pool of workers using all of the available ones.
    
    # Start all the processors!             
    ResultData = [pool.apply_async(RunAspenPlusReactor, args=(JobID, PointsList[JobID])) for JobID in range(NumAspenCopiesOpen)]    
    
    # Closes the parallel worker pool        
    pool.close()

    # This waits until all the above jobs are finished before continuing on
    pool.join()        
                   
    # Stop the timer. We are done!                   
    TimeFinish = time.perf_counter()            
    print(f"Finished in {TimeFinish - TimeStart:0.2f} Seconds.")
        

