########
# (c) Thomas A. Adams II
# December 2020
# From Bonus Tutorial 2 in the book
# LEARN ASPEN PLUS IN 24 HOURS, McGraw-Hill Education
#
# This code does Particle Swarm Optimization on Aspen Plus simulations using parallel processing
#
# This version has error checking. It uses the following strategy: 
#
#   Each process has one copy of Aspen Plus open
#   Each simulation run uses the results of the previous run as the initial guesses, so no reinitilzation is used.
#   If any errors occurred, it reinitializes and tries again
#   If still any errors occurred after reinitializing, it closes Aspen PLus and opens a new copy and tries again
#   If that still doesn't work, it records a 0 for all variables, gives up, and moves on.
#
# Each PSO iteration generates a new pool of workers. This means that Aspen Plus is re-dispatched each iteration.
# The number of particles should be about 10 times the number of processes you use
# You should use 1/2 to 1/3 of the number of logical cores you have available as processes in this code
#
# USER PARAMETERS TO SET
# Users should change these four numbers found near the top of the main function:
#
#   MaxPSOIteration = 100   --> Maximum number of PSO iterations to run in case convergence is not reached in a timely manner
#   Consensus = 0.005       --> Indicates when you can stop early because the particles are all close togeher. The smaller this number (must be positive), the more stringent the requirement and the longer the run. 
#                           --> A consensus of 0.005 means that it should stop when all particles are within an N-dimensional box centred around the global best where the length of each dimension of the box is 0.5% of the range of that dimension. 
#   NumAspenCopiesOpen = 4  --> This should be between 1 and the number of processors your computer has, but no more than 60     
#   PointsPerProcessor = 8  --> This is how many particles points you want *per* processor, *per* PSO iteration. The total number of points explored is NumAspenCopiesOpen * PointsPerProcessor
#                           --> Typically, NumAspenCopiesOpen * PointsPerProcessor should be in the 20-40 range but for high dimensional problems you can use more, or when you have many processors handy. 
#
# HOW TO RUN
# In the windows command prompt, type: 
#
#     python Part3.py
#
# If you want to save the results to a file instead of seeing them on the screen, instead type 
#
#     python Part3.py > filename.txt
#
# USEFUL FUNCTIONS
# CRTL + C or CTRL + Break during the run will stop it, for example, if it is taking too long. Expect betwen 0.15 and 0.5 seconds per point visited. 
# If you break it early or something goes wrong, you may have many open copies of Aspen Plus in background. If it is too tedious to close them with the Windows Task Manager you can use the following command prompt command:
#
#     taskkill /IM "AspenPlus.exe" /f   
#
# If you find there are too many temporary Aspen plus files lying about (those that start with underscore) you can type this at the command prompt:
#
#     del _*.*
#
# Just be sure to include that _ or else you will delete all files!!!
#
########

import os                            # This connects to your operating system, and lets you make operating system calls, like get paths to directories, change directories, etc   
import win32com.client as Windows    # This is a package that lets you call Windows programs
import multiprocessing as mp         # This is the multiprocessing package
import random                        # This lets us call the random number generator
import time                          # This has the timer function in it
import math                          # This has the square root function


# This runs the Aspen Sim for the given T and P, and returns the result. Does not do error checking.
def RunAspenPlusReactor(JobID, Points):
    
    # Create the Aspen Plus Object     
    AspenPlus = Windows.Dispatch('Apwn.Document')
    
    # These are variables that keep track of how many times certain errors happened 
    reinits = 0
    redispatches = 0
    varfails = 0
    finalfails = 0    
    
    # Load the simulation from the file, and fail if it doesn't work the first time
    # If it doesn't work the first time, usually this means the file name is wrong, or the license checkout failed.
    try: 
        AspenPlus.InitFromArchive2(os.path.abspath('Tutorial B02.bkp'))
    except:
        print(f"Could not load the file #{JobID}. Ensure that it is in the same directory as the python script. Also, ensure that you are connected to a license server and that you have Aspen Plus licenses available.")
        return ([])
    
    # Create the empty list for storing results
    Data = list()

    # Now, run a simulation for each point we were given
    for point in range(len(Points)):
            
        # This keeps track of how many times we attempted to run the simulation for this point 
        attempt = 0
        
        # Default values for the results are stored here, in case of failure
        Q = float('Inf')
        CO = 0
        H2 = 0
        
        # Run The Simulation. We allow 3 attempts before giving up.
        while (3 > attempt):
            
            # If our first attempt failed, try re-initializaing the simulation and try again
            # This could be just because the initial guess (the previous run) was poor
            if (1 == attempt):                
                print("Attempting to reinitialize ", JobID)
                try:
                    # Reinitialize the simulation
                    AspenPlus.Engine.Reinit()
                    reinits = reinits + 1
                except:
                    print("Could not reinitialize ", JobID)
                    attempt = attempt + 1                    
                else: 
                    print("Reinitialize successful ", JobID)                   
            
            # If our second attempt failed, try relaunching Aspen Plus entirely             
            if (2 == attempt):                
                print("Attempting to redispatch Aspen Plus", JobID)                                
                try:
                    AspenPlus = Windows.Dispatch('Apwn.Document')
                    redispatches = redispatches + 1
                except:
                    print("Could not dispatch on reopen ", JobID)
                    break     
                else:
                    print("Dispatch successful ", JobID)                    
                
                # At this point, the dispatch worked, so try to open the file again.                 
                try:                                 
                    AspenPlus.InitFromArchive2(os.path.abspath('Tutorial B02.bkp'))
                except:
                    print("Could not initialize on reopen ", JobID)
                    break                     
                else:
                    print("Initialize on reopen successful ", JobID)
            
            # At this point, we have an open and working file.
            # Set the T and P variables for the current point as the T and P of the reactor in the simulation                                   
            AspenPlus.Tree.FindNode('\Data\Blocks\REACTOR\Input\TEMP').Value = Points[point][0] # Temperature in Celcius
            AspenPlus.Tree.FindNode('\Data\Blocks\REACTOR\Input\PRES').Value = Points[point][1] # Pressure in bar                        
            
            # Run the simulation                                           
            try:
                AspenPlus.Engine.Run2()
            except:
                # If it did not run, print an error message
                print("Aspen Run Error in jobID ", JobID, " on attempt" , attempt)
                # Increment the attempt counter, so can try again in the next while loop iteration
                attempt = attempt + 1                                    
            else:
                # The simulation was successful. Let's record that it worked this time for later.
                worked = 1                            
                
                # Sometimes the simulation can be successful but some of the simulation varibles are missing
                # We need to check for that here.
                try:                
                    Q = AspenPlus.Tree.FindNode('\Data\Blocks\REACTOR\Output\QCALC').Value   # Units of Gcal/hr
                except:
                    # Even though this simulation worked, there is some problem in Aspen Plus that prevents us 
                    # From seeing the variable. So We can set a default variable and indicate that it did not work.
                    Q = float("inf")
                    worked = 0
                    varfails = varfails + 1
                    
                # We need to check against each variable that we want to keep. A little tedious but important for robustness     
                try: 
                    CO = AspenPlus.Tree.FindNode('\Data\Streams\PRODUCTS\Output\MASSFLOW\MIXED\CO').Value  # Units of kg/hr
                except:
                    CO = 0
                    worked = 0
                    varfails = varfails + 1
                    
                # We need to check against each variable that we want to keep. A little tedious but important for robustness     
                try: 
                    H2 = AspenPlus.Tree.FindNode('\Data\Streams\PRODUCTS\Output\MASSFLOW\MIXED\H2').Value # Units of kg/hr
                except:
                    H2 = 0
                    worked = 0
                    varfails = varfails + 1
                    
                # If none of the variable pulls failed, then we can break from our while loop because the point was successful     
                if (1 == worked):                
                    break
                else:
                    # Otherwise we just try again
                    attempt = attempt + 1    
         
        # This checks to see if the previous while loop exited with a simulation run and all output variables obtained.
        # If not, we just keep track of the number of times this happened for convenience.             
        if (2 < attempt):
            finalfails = finalfails + 1                                            
        
        
        # Compute the objective function using:
        # +$16.00  per Gcal of heating duty (cost of natural gas)
        # +$ 2.00  per GCal of cooling duty (cost of cooling system)
        # -$ 0.1   per kg of H2 (value to us of the H2 we produce)
        # -$ 0.005 per kg of CO produced (value to us)
        # Need about $10 per hour per compressor stage for recompression to 51 bar after the reactor downstream
        # Use 50 * sqrt((50-P)/5) as cost of low pressure
        # Units on OBJ are $/hr
                
        Obj = -0.100 * H2 - 0.005 * CO + 50 * math.sqrt( (51 - Points[point][1]) / 5 ) 
        if (Q < 0): # If cooling duty
            Obj = Obj + 2 * -Q
        else: # heating duty
            Obj = Obj + 16 * Q                        
                
        # Store and print the data. Although I am not doing anything with the Data object, you could do something with it if you wanted.                                 
        print(f"Job ID: {JobID}, {Points[point][0]:0.4f}, {Points[point][1]:0.4f}, {Q:0.4f}, {CO:0.4f}, {Obj:0.4f}")
        
        # Store the objective function result. In this case we want to have 0 heat duty so we minimize Q^2
        Data.append(Obj)      
    
    
    # Now, going through the data and determine the best one out of this pool.
    
    # Close this out and spit out the data.  
    print("JobID: ", JobID, "Finished with this worker. Completed ", len(Data), " points, with ", reinits, " reinitializes, ", redispatches, " redispatches, ", varfails, " variable pull fails, and ", finalfails, " complete failures.")
               
    return (JobID, Data)                
        
if __name__ == '__main__': 
     
    #########################
    # USER CHANGE THIS PART #
    # (see comments above)  #
    #########################
    MaxPSOIteration = 100   
    Consensus = 0.005        
    NumAspenCopiesOpen = 4     
    PointsPerProcessor = 8   
    #########################
    #########################
    
    N = 2 # This is the number of dimensions in the problem. In our case, 2 (T and P of the reactor)
    Np = NumAspenCopiesOpen * PointsPerProcessor # These are the number of particles to use
    minrange = [600, 2]     # This is the minimum permitted values of each of your decision variables
    maxrange = [1000, 50]    # This is the maximum permitted values of each of your decision variables

    # Initialize the location and velocity of each particle using Latin Hypercube Sampling.
    x = [ [] for d in range(N) ] # Creates an empty list of locations for each particle 
    v = [ [] for d in range(N) ] # Creates an empty list of velocities for each particle

    # For each particle, set its initial position and velocity.
    for d in range(N):
        for i in range(Np):     
            x[d].append(minrange[d] + (maxrange[d]-minrange[d]) * (i/(Np-1)))        # Creates a point on the grid
            v[d].append(0.5 * (random.random() - 0.5) * (maxrange[d]-minrange[d]))   # Velocity is +/- 25% of the range, either positive or negative
        if (0 < d):
            random.shuffle(x[d])  # Randomizes each dimension (no need to do this on the first dimension)                                         

    # PSO Tuning Parameters. These values are classically chosen.    
    w1 = 2.8 # classic parameters
    w2 = 1.3 # classic parameters
    chi = -2 / (2-w1-w2-math.sqrt((w1+w2)**2 - 4*(w1+w2)))    
    
    # start the timer
    TimeStart = time.perf_counter()

    # Remember the personal and global bests of each particle. Start with the current point and Infinite
    pbestx = x.copy();
    pbestf = [float("inf")] * Np
    gbestf = float("inf");  # This assumes the objective is to minimize the function.
    gbestx = [float("NaN")] * N
    prevx = list();
    
    # Check to make sure that the user did not specify the wrong number of processors
    if (0 > NumAspenCopiesOpen or mp.cpu_count() < NumAspenCopiesOpen):
        NumAspenCopiesOpen = mp.cpu_count()
    if (60 < NumAspenCopiesOpen):
        NumAspenCopiesOpen = 60 
        
    # Let's see how many processors we have (this is logical, not physical)
    print("Number of processors available: ", mp.cpu_count())
    print("Number of processes used: ", NumAspenCopiesOpen)        
    
    # Start the PSO. We iterate until either the maximum has been reached or a consensus has been reached.    
    for i in range(MaxPSOIteration):
    
        # Create the pool of workers for this iteration
        pool = mp.Pool(NumAspenCopiesOpen)             
    
        # Remember the previous iteration 
        if (0 < i):
            prevx.clear() # only need to clear it after the 1st iteration since it starts empty
        prevx = x.copy() 
        
        # Create the job lists of the points that each processor will need to process
        PointsList = []        
        for a in range(NumAspenCopiesOpen):
            for c in range(PointsPerProcessor):
                if (1 > c):
                    PointsList.append(list())
                PointsList[a].append( [ x[d][c + a * PointsPerProcessor] for d in range(N) ] )                                            
        
        # Evaluate the function for each particle 
        ResultData = [pool.apply_async(RunAspenPlusReactor, args=(JobID, PointsList[JobID])) for JobID in range(NumAspenCopiesOpen)]    
        
        # The ResultData does not contain our results in a way that we can access them.
        # We will use this to access the results of each function
        Results = [Result.get()[1] for Result in ResultData]    
        
        # Closes the parallel worker pool
        pool.close()
    
        # This waits until all the above jobs are finished before continuing on
        pool.join()   
    
        # Check each point, and update personal and global bests accordingly
        for a in range(NumAspenCopiesOpen):
            for c in range(PointsPerProcessor):
                k = c + a * PointsPerProcessor
                if (gbestf > Results[a][c]): # update the global best
                    gbestf = Results[a][c]
                    for d in range(N):
                        gbestx[d] = x[d][k]
                if (pbestf[k] > Results[a][c]): # update the personal best
                    pbestf[k] = Results[a][c]
                    for d in range(N):
                        pbestx[d][k] = x[d][k]
        
        # update velocities of each particle and move them accordingly.
        for k in range(Np):
            for d in range(N):                
                # update the velocity (at current dimension)
                v[d][k] = chi * ( v[d][k] + w1 * random.random() * (pbestx[d][k] - x[d][k]) + w2 * random.random() * (gbestx[d] - x[d][k])) 
                # move the particle (at current dimension)
                x[d][k] = x[d][k] + v[d][k] 
                # enforce the bounds (using sticky bounds, leaving velocities intact)
                if (minrange[d] > x[d][k]):
                    x[d][k] = minrange[d]
                if (maxrange[d] < x[d][k]):
                    x[d][k] = maxrange[d]
          
        # determine a normalized cloud size, to check for early exit.
        # We can exit early when all points are within 0.5*Consensus % of the global best        

        stop = True
        for k in range(Np):
            for d in range(N):
                if (gbestx[d] + 0.5*Consensus*(maxrange[d]-minrange[d]) < x[d][k] or gbestx[d] - 0.5*Consensus*(maxrange[d]-minrange[d]) > x[d][k]):
                    stop = False     
                    break
            if (False == stop):
                break
        
        if (stop): # all particles are within the box
            print(f"Terminating early after {i} iterations since all particles have reached a consensus within {Consensus}")
            break


    #### END ITERATIONS
    
    # Spit out the results
    TimeFinish = time.perf_counter()
    print(f"Finished with the parallel computing part in {TimeFinish - TimeStart:0.2f} Seconds.")
    
    # Determine the consensus:
    Cons = 0
    for k in range(Np):
        for d in range(N):
            if (abs(gbestx[d] - x[d][k]) / (maxrange[d]-minrange[d]) > Cons):
                Cons = abs(gbestx[d] - x[d][k]) / (maxrange[d]-minrange[d]) 
    Cons = Cons * 2                                                  
    print(f"Final consensus was: {Cons:0.4f}")
    print(f"Global best was {gbestf:0.4f} at ")
    for d in range(N):
        print(f"x{d}: {gbestx[d]:0.4f}")

    print(f"Program complete. {(i+1) * Np} Aspen Plus calls completed. Approximately {(TimeFinish - TimeStart)/((i+1)*Np):0.3f} Seconds per call." )
        

